@@ -8,15 +8,13 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2009 by Bradford W. Mott and the Stella team
+// Copyright (c) 1995-2014 by Bradford W. Mott, Stephen Anthony
+// and the Stella Team
 //
-// See the file "license" for information on usage and redistribution of
+// See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
 //
-// $Id: DataGridWidget.cxx 1755 2009-06-08 13:13:05Z stephena $
-//
-//   Based on code from ScummVM - Scumm Interpreter
-//   Copyright (C) 2002-2004 The ScummVM project
+// $Id: DataGridWidget.cxx 2838 2014-01-17 23:34:03Z stephena $
 //============================================================================
 
 #include "OSystem.hxx"
@@ -25,12 +23,15 @@
 #include "Debugger.hxx"
 #include "FrameBuffer.hxx"
 #include "DataGridWidget.hxx"
+#include "DataGridOpsWidget.hxx"
 #include "RamWidget.hxx"
+#include "ScrollBarWidget.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 DataGridWidget::DataGridWidget(GuiObject* boss, const GUI::Font& font,
                                int x, int y, int cols, int rows,
-                               int colchars, int bits, BaseFormat base,
+                               int colchars, int bits,
+                               Common::Base::Format base,
                                bool useScrollbar)
   : EditableWidget(boss, font, x, y,
                    cols*(colchars * font.getMaxCharWidth() + 8) + 1,
@@ -44,25 +45,29 @@ DataGridWidget::DataGridWidget(GuiObject* boss, const GUI::Font& font,
     _bits(bits),
     _base(base),
     _selectedItem(0),
+    _currentKeyDown(KBDK_UNKNOWN),
     _opsWidget(NULL),
     _scrollBar(NULL)
 {
   _flags = WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS |
            WIDGET_WANTS_RAWDATA;
-  _type = kDataGridWidget;
   _editMode = false;
 
-  _currentKeyDown = 0;
-
   // The item is selected, thus _bgcolor is used to draw the caret and
   // _textcolorhi to erase it
   _caretInverse = true;
 
-  // Make sure hilite list contains all false values
+  // Make sure all lists contain some default values
   _hiliteList.clear();
   int size = _rows * _cols;
   while(size--)
+  {
+    _addrList.push_back(0);
+    _valueList.push_back(0);
+    _valueStringList.push_back("");
+    _changedList.push_back(0);
     _hiliteList.push_back(false);
+  }
 
   // Set lower and upper bounds to sane values
   setRange(0, 1 << bits);
@@ -98,7 +103,6 @@ cerr << "alist.size() = "     << alist.size()
   assert(size == _rows * _cols);
 
   _addrList.clear();
-  _addrStringList.clear();
   _valueList.clear();
   _valueStringList.clear();
   _changedList.clear();
@@ -110,10 +114,7 @@ cerr << "alist.size() = "     << alist.size()
   // An efficiency thing
   string temp;
   for(int i = 0; i < size; ++i)
-  {
-    temp = instance().debugger().valueToString(_valueList[i], _base);
-    _valueStringList.push_back(temp);
-  }
+    _valueStringList.push_back(Common::Base::toString(_valueList[i], _base));
 
 /*
 cerr << "_addrList.size() = "     << _addrList.size()
@@ -125,7 +126,7 @@ cerr << "_addrList.size() = "     << _addrList.size()
   _editMode = false;
 
   // Send item selected signal for starting with cell 0
-  sendCommand(kDGSelectionChangedCmd, _selectedItem, _id);
+  sendCommand(DataGridWidget::kSelectionChangedCmd, _selectedItem, _id);
 
   setDirty(); draw();
 }
@@ -143,6 +144,20 @@ void DataGridWidget::setList(int a, int v, bool c)
   setList(alist, vlist, changed);
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void DataGridWidget::setList(int a, int v)
+{
+  IntArray alist, vlist;
+  BoolArray changed;
+
+  alist.push_back(a);
+  vlist.push_back(v);
+  bool diff = _addrList.size() == 1 ? getSelectedValue() != v : false;
+  changed.push_back(diff);
+
+  setList(alist, vlist, changed);
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void DataGridWidget::setHiliteList(const BoolArray& hilitelist)
 {
@@ -178,13 +193,13 @@ void DataGridWidget::setValue(int position, int value, bool changed)
   if(position >= 0 && uInt32(position) < _valueList.size())
   {
     // Correctly format the data for viewing
-    _editString = instance().debugger().valueToString(value, _base);
+    _editString = Common::Base::toString(value, _base);
 
     _valueStringList[position] = _editString;
     _changedList[position] = changed;
     _valueList[position] = value;
 
-    sendCommand(kDGItemDataChangedCmd, position, _id);
+    sendCommand(DataGridWidget::kItemDataChangedCmd, position, _id);
 
     setDirty(); draw();
   }
@@ -218,7 +233,7 @@ void DataGridWidget::handleMouseDown(int x, int y, int button, int clickCount)
     _currentRow = _selectedItem / _cols;
     _currentCol = _selectedItem - (_currentRow * _cols);
 
-    sendCommand(kDGSelectionChangedCmd, _selectedItem, _id);
+    sendCommand(DataGridWidget::kSelectionChangedCmd, _selectedItem, _id);
     setDirty(); draw();
   }
 }
@@ -230,7 +245,7 @@ void DataGridWidget::handleMouseUp(int x, int y, int button, int clickCount)
   // send the double click command
   if (clickCount == 2 && (_selectedItem == findItem(x, y)))
   {
-    sendCommand(kDGItemDoubleClickedCmd, _selectedItem, _id);
+    sendCommand(DataGridWidget::kItemDoubleClickedCmd, _selectedItem, _id);
 
     // Start edit mode
     if(_editable && !_editMode)
@@ -241,7 +256,15 @@ void DataGridWidget::handleMouseUp(int x, int y, int button, int clickCount)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void DataGridWidget::handleMouseWheel(int x, int y, int direction)
 {
-  _scrollBar->handleMouseWheel(x, y, direction);
+  if(_scrollBar)
+    _scrollBar->handleMouseWheel(x, y, direction);
+  else if(_editable)
+  {
+    if(direction > 0)
+      decrementCell();
+    else if(direction < 0)
+      incrementCell();
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -257,11 +280,11 @@ int DataGridWidget::findItem(int x, int y)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
+bool DataGridWidget::handleKeyDown(StellaKey key, StellaMod mod, char ascii)
 {
   // Ignore all mod keys
-  if(instance().eventHandler().kbdControl(modifiers) ||
-     instance().eventHandler().kbdAlt(modifiers))
+  if(instance().eventHandler().kbdControl(mod) ||
+     instance().eventHandler().kbdAlt(mod))
     return true;
 
   bool handled = true;
@@ -270,17 +293,17 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
   if (_editMode)
   {
     // Class EditableWidget handles all text editing related key presses for us
-    handled = EditableWidget::handleKeyDown(ascii, keycode, modifiers);
+    handled = EditableWidget::handleKeyDown(key, mod, ascii);
     if(handled)
       setDirty(); draw();
   }
   else
   {
     // not editmode
-    switch(ascii)
+    switch(key)
     {
-      case '\n':  // enter/return
-      case '\r':
+      case KBDK_RETURN:
+      case KBDK_KP_ENTER:
         if (_currentRow >= 0 && _currentCol >= 0)
         {
           dirty = true;
@@ -289,7 +312,7 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+17:  // up arrow
+      case KBDK_UP:
         if (_currentRow > 0)
         {
           _currentRow--;
@@ -303,7 +326,7 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+18:  // down arrow
+      case KBDK_DOWN:
         if (_currentRow < (int) _rows - 1)
         {
           _currentRow++;
@@ -317,7 +340,7 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+20:  // left arrow
+      case KBDK_LEFT:
         if (_currentCol > 0)
         {
           _currentCol--;
@@ -331,7 +354,7 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+19:  // right arrow
+      case KBDK_RIGHT:
         if (_currentCol < (int) _cols - 1)
         {
           _currentCol++;
@@ -345,8 +368,8 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+24:  // pageup
-        if(instance().eventHandler().kbdShift(modifiers) && _scrollBar)
+      case KBDK_PAGEUP:
+        if(instance().eventHandler().kbdShift(mod) && _scrollBar)
           handleMouseWheel(0, 0, -1);
         else if (_currentRow > 0)
         {
@@ -355,8 +378,8 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+25:	// pagedown
-        if(instance().eventHandler().kbdShift(modifiers) && _scrollBar)
+      case KBDK_PAGEDOWN:
+        if(instance().eventHandler().kbdShift(mod) && _scrollBar)
           handleMouseWheel(0, 0, +1);
         else if (_currentRow < (int) _rows - 1)
         {
@@ -365,7 +388,7 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+22:  // home
+      case KBDK_HOME:
         if (_currentCol > 0)
         {
           _currentCol = 0;
@@ -373,7 +396,7 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 256+23:  // end
+      case KBDK_END:
         if (_currentCol < (int) _cols - 1)
         {
           _currentCol = _cols - 1;
@@ -381,48 +404,58 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
         }
         break;
 
-      case 'n': // negate
+      case KBDK_n: // negate
         if(_editable)
           negateCell();
         break;
 
-      case 'i': // invert
-      case '!':
-        if(_editable)
-          invertCell();
-        break;
-
-      case '-': // decrement
-        if(_editable)
-          decrementCell();
-        break;
-
-      case '+': // increment
-      case '=':
-        if(_editable)
-          incrementCell();
-        break;
-
-      case '<': // shift left
-      case ',':
-        if(_editable)
-          lshiftCell();
-        break;
-
-      case '>': // shift right
-      case '.':
-        if(_editable)
-          rshiftCell();
-        break;
-
-      case 'z': // zero
-        if(_editable)
-          zeroCell();
-        break;
-
       default:
         handled = false;
     }
+    if(!handled)
+    {
+      handled = true;
+
+      switch(ascii)
+      {
+        case 'i': // invert
+        case '!':
+          if(_editable)
+            invertCell();
+          break;
+
+        case '-': // decrement
+          if(_editable)
+            decrementCell();
+          break;
+
+        case '+': // increment
+        case '=':
+          if(_editable)
+            incrementCell();
+          break;
+
+        case '<': // shift left
+        case ',':
+          if(_editable)
+            lshiftCell();
+          break;
+
+        case '>': // shift right
+        case '.':
+          if(_editable)
+            rshiftCell();
+          break;
+
+        case 'z': // zero
+          if(_editable)
+            zeroCell();
+          break;
+
+        default:
+          handled = false;
+      }
+    }
   }
 
   if (dirty)
@@ -431,20 +464,20 @@ bool DataGridWidget::handleKeyDown(int ascii, int keycode, int modifiers)
     _selectedItem = _currentRow*_cols + _currentCol;
 
     if(_selectedItem != oldItem)
-      sendCommand(kDGSelectionChangedCmd, _selectedItem, _id);
+      sendCommand(DataGridWidget::kSelectionChangedCmd, _selectedItem, _id);
 
     setDirty(); draw();
   }
 
-  _currentKeyDown = keycode;
+  _currentKeyDown = key;
   return handled;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool DataGridWidget::handleKeyUp(int ascii, int keycode, int modifiers)
+bool DataGridWidget::handleKeyUp(StellaKey key, StellaMod mod, char ascii)
 {
-  if (keycode == _currentKeyDown)
-    _currentKeyDown = 0;
+  if (key == _currentKeyDown)
+    _currentKeyDown = KBDK_UNKNOWN;
   return true;
 }
 
@@ -516,7 +549,6 @@ void DataGridWidget::drawWidget(bool hilite)
 //cerr << "DataGridWidget::drawWidget\n";
   FBSurface& s = _boss->dialog().surface();
   int row, col, deltax;
-  string buffer;
 
   // Draw the internal grid and labels
   int linewidth = _cols * _colWidth;
@@ -542,16 +574,14 @@ void DataGridWidget::drawWidget(bool hilite)
 
       if (_selectedItem == pos && _editMode)
       {
-        buffer = _editString;
         adjustOffset();
         deltax = -_editScrollOffset;
 
-        s.drawString(_font, buffer, x, y, _colWidth, kTextColor,
+        s.drawString(_font, _editString, x, y, _colWidth, kTextColor,
                      kTextAlignLeft, deltax, false);
       }
       else
       {
-        buffer = _valueStringList[pos];
         deltax = 0;
 
         uInt32 color = kTextColor;
@@ -567,7 +597,7 @@ void DataGridWidget::drawWidget(bool hilite)
         else if(_hiliteList[pos])
           color = kDbgColorHi;
 
-        s.drawString(_font, buffer, x, y, _colWidth, color);
+        s.drawString(_font, _valueStringList[pos], x, y, _colWidth, color);
       }
     }
   }
@@ -595,13 +625,19 @@ GUI::Rect DataGridWidget::getEditRect() const
   return r;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int DataGridWidget::getWidth() const
+{
+  return _w + (_scrollBar ? kScrollBarWidth : 0);
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void DataGridWidget::startEditMode()
 {
   if (_editable && !_editMode && _selectedItem >= 0)
   {
     _editMode = true;
-    setEditString("");  // Erase current entry when starting editing
+    setText("", true );  // Erase current entry when starting editing
   }
 }
 
@@ -611,10 +647,33 @@ void DataGridWidget::endEditMode()
   if (!_editMode)
     return;
 
-  // send a message that editing finished with a return/enter key press
   _editMode = false;
 
   // Update the both the string representation and the real data
+  if(_editString.size() > 0 && !(_editString[0] == '$' ||
+        _editString[0] == '#' || _editString[0] == '\\'))
+  {
+    switch(_base)
+    {
+      case Common::Base::F_16:
+      case Common::Base::F_16_1:
+      case Common::Base::F_16_2:
+      case Common::Base::F_16_4:
+      case Common::Base::F_16_8:
+        _editString.insert(0, 1, '$');
+        break;
+      case Common::Base::F_2:
+      case Common::Base::F_2_8:
+      case Common::Base::F_2_16:
+        _editString.insert(0, 1, '\\');
+        break;
+      case Common::Base::F_10:
+        _editString.insert(0, 1, '#');
+        break;
+      case Common::Base::F_DEFAULT:
+        break;
+    }
+  }
   int value = instance().debugger().stringToValue(_editString);
   if(value < _lowerBound || value >= _upperBound)
   {
@@ -636,16 +695,34 @@ void DataGridWidget::abortEditMode()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool DataGridWidget::tryInsertChar(char c, int pos)
 {
-  // Not sure how efficient this is, or should we even care?
+  // Input is very strict here, to eliminate time-consuming error checking
+  // elsewhere, and includes the following restrictions:
+  //    Cannot contain spaces
+  //    Starts with leading specifier ($, #, \), or with a base character
+  //    Only one specifier is allowed
+  //    If starting with a specifier, only allow numbers applicable to that
+  //    base to follow
+
   c = tolower(c);
-  if((c >= '0' && c <= '9') || (c >= 'a' && c <= 'f') ||
-     c == '%' || c == '#' || c == '$')
-  {
+  bool isBin = c == '0' || c == '1',
+       isDec = c >= '0' && c <= '9',
+       isHex = isDec || (c >= 'a' && c <= 'f'),
+       isOp = c == '$' || c == '#' || c == '\\',
+       insert = false;
+
+  if(BSPF_startsWithIgnoreCase(_editString, "$"))
+    insert = isHex && pos > 0;
+  else if(BSPF_startsWithIgnoreCase(_editString, "#"))
+    insert = isDec && pos > 0;
+  else if(BSPF_startsWithIgnoreCase(_editString, "\\"))
+    insert = isBin && pos > 0;
+  else 
+    insert = isHex || isDec || isBin || isOp;
+
+  if(insert)
     _editString.insert(pos, 1, c);
-    return true;
-  }
-  else
-    return false;
+
+  return insert;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
